import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import javax.swing.*;
import java.lang.Math;
import java.util.Random;

/** 
 * An interface to the robot information.  Given some script file and
 * map file (both will be provided) you will be able to get status
 * information at each step the robot takes.
 * @author Matt Hoffman
 * @version 1.0
 ***/
public class RobotModel
{
    private JFrame mFrame;
    private BufferedReader mScript;
    private GridMap mMap;
    private int mCurrX = 0;
    private int mCurrY = 0;
    private float mCurrR = 0.0f;
    private Random mRandom;

    /** The initial x-position for the robot. NOTE: this will be
     * removed for the second half of the project. */
    public final int initialX; 

    /** The initial x-position for the robot. NOTE: this will be
     * removed for the second half of the project. */
    public final int initialY;
    
    /**
     * Construct a robot model object given the path to some script
     * file, and some map file. If the map-file is null, this will
     * give you an open area with no walls to test your code. All file
     * locations should be relative to the current working directory,
     * ie the directory from which the program is launched.
     * @param script The location of script file.
     * @param map The location of the map file.
     ***/
    public RobotModel (String script, String map)
    {
        mRandom = new Random(System.currentTimeMillis());
        
        // try to open our file and then discard everything up to the
        // first scan; any errors that occur should be fatal.
        try {
            mScript = new BufferedReader(new FileReader(script));

            // check the version
            if(!(mScript.readLine()).startsWith("version: 1")) 
                throw new Exception ("Invalid file version! Either " +
                        "your file is wrong, or you should upgrade.");

            String[] strings = (mScript.readLine()).split("\\s+");

            if(!strings[0].contentEquals("init"))
                throw new Exception("Invalid file.");
            
            mCurrX = Integer.parseInt(strings[1]);
            mCurrY = Integer.parseInt(strings[2]);
            mCurrR = Float.parseFloat(strings[3]);
        } 
        catch (Exception e) {
            System.out.println("" + e);
            System.exit(-1);
        }

        // XXX: set the initial position variables; this should be
        // removed for the second half of the project
        initialX = mCurrX;
        initialY = mCurrY;

        mMap = new GridMap(map, mCurrX, mCurrY);
        mFrame = new JFrame("Display");
        mFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        mFrame.getContentPane().add(mMap);
        mFrame.pack();
        mFrame.setVisible(true);

        System.out.println("Initially at (" + mCurrX + "," +
                                              mCurrY + ")");
        System.out.println("--------------------------------");
    }

    /** Generate a gaussian with mean 0.0 and variance 1.0. */
    public double randGaussian() { return randGaussian(0.0, 1.0); }

    /**
     * Generate a gaussian with mean m and variance s^2.
     * @param m The mean of the distribution.
     * @param s The standard deviation.
     ***/
    public double randGaussian(double m, double s)
    {
        return (s * mRandom.nextGaussian() + m);
    }
    
    /**
     * Return whether or not the position is occupied. This will tell
     * you whether the robot can move there or not, and whether or not
     * you should put particles there.
     * @param x The x position.
     * @param y The y position.
     ***/
    public boolean occupied(int x, int y) 
    { 
        if (y>=0 && x>=0 && y<mMap.maxY && x<mMap.maxX)
            return mMap.map[y][x];
        else
            return false;
    }

    /** Schedule the map to be redrawn. */
    public void redraw()
    {
        mMap.repaint(0, 0, 0, mMap.getWidth(), mMap.getHeight());
    }

    /** Add a particle to the map. */
    public void addParticle(int x, int y) 
    {
        mMap.addParticle(x,y);
    }

    /** 
     * Get the next status-update for the robot, ie its movement and
     * sensor information. This method will return null when the robot
     * is done moving.
     ***/
    public Status getNext()
    {
        try {
            String line = mScript.readLine();
            if (line != null)
            {
                // make a new status object, which will split the line
                // into d/i/f entries
                Status status = new Status(line);

                // update our current knowledge of the robot's
                // position
                mCurrR += status.initialRotation;
                mCurrX += (int) (Math.cos(mCurrR + Math.PI/2) *
                      status.distance - .5);
                mCurrY += (int) -(Math.sin(mCurrR + Math.PI/2) *
                      status.distance + .5);
                mCurrR += status.finalRotation;

                // set the robot's new position and return this status
                mMap.clearParticles();
                mMap.setPosition(mCurrX, mCurrY);
                return status;
            }
        }
        catch (Exception e) {
            System.out.println("" + e);
            System.exit(-1);
        }

        return null;
    }
};
